module:    Simple-Printer-Implementation
author:    Matthias Hlzl (tc@gauss.muc.de)
copyright: Copyright (C) 1998 Matthias Hlzl
           Copyright (C) 1998 Way Forward Technologies
version:   0.03 09 Dec 1998
synopsis:  This Module implements some simple output methods.

// Copyright.
// =========

// The Simple-Printer module.
//
// Copyright (C) 1998 Matthias Hlzl.
// Copyright (C) 1998 Way Forward Technologies.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston,
// MA 02111-1307  USA
//
// If you need to receive this program under another license contact
// the author (tc@gauss.muc.de).


// The method SAY takes a stream, a format string and optional
// arguments. It formats its arguments on the stream (if any arguments
// are supplied) and forces the stream to output.  The result is #t if
// any output was written, #f if the stream was only forced.
//
define method say (stream :: <stream>, #rest arguments)
 => wrote-something? :: <boolean>;
  let wrote-something? = #t;
  if (arguments.size = 1)
    apply(write, stream, arguments);
  elseif (arguments.size >= 2)
    apply(format, stream, arguments);
  else
    wrote-something? := #f;
  end if;
  stream.force-output;
  wrote-something?;
end method say;

// Instead of the stream we can simply specify #t or #f to print to
// standard output or a string, respectively.
//
define method say (stream == #t, #rest arguments)
 => wrote-something? :: <boolean>;
  apply(say, *standard-output*, arguments);
end method say;

define method say (stream == #f, #rest arguments)
 => formatted-string :: <byte-string>;
  let out-stream = make(<byte-string-stream>,
			contents: make(<byte-string>, size: 200),
			direction: #"output");
  apply(say, out-stream, arguments);
  out-stream.stream-contents;
end method say;

// SAY-LINE is used to display a message taking a whole line.
//
define method say-line (stream :: <stream>, #rest arguments) => ();
  apply(say, stream, arguments);
  write-line(stream, "");
  stream.force-output;
end method say-line;

define method say-line (stream == #t, #rest arguments) => ();
  apply(say-line, *standard-output*, arguments);
end method say-line;

define method say-line (stream == #f, #rest arguments)
 => formatted-string :: <string>;
  concatenate(apply(say, #f, arguments), "\n");
end method say-line;

// SAY-DOTS displays its arguments with three trailing dots.
//
define method say-dots (stream :: <stream>, #rest arguments)
 => printed-something? :: <boolean>;
  if (arguments.size > 0)
    let text = apply(say-dots, #f, arguments);
    say(stream, text);
  else
    #f;
  end if;
end method say-dots;

define method say-dots (stream == #t, #rest arguments)
 => printed-something? :: <boolean>;
  apply(say-dots, *standard-output*, arguments);
end method say-dots;

define method say-dots (stream == #f, #rest arguments)
 => formatted-string :: <string>;
  concatenate(apply(say, #f, arguments), "...");
end method say-dots;

// The method SAY-UNDERLINED prints its arguments to the stream and
// underlines this output.
//
define method say-underlined (stream :: <stream>,
			      underline-char :: <character>,
			      #rest arguments)
 => printed-something? :: <boolean>;
  if (arguments.size > 0)
    let (formatted-string, underlining)
      = apply(say-underlined, #f, underline-char, arguments);
    say-line(stream, formatted-string);
    say-line(stream, underlining);
  else
    say(stream);
  end if;
end method say-underlined;

define method say-underlined (stream == #t,
			      underline-char :: <character>,
			      #rest arguments)
 => printed-something? :: <boolean>;
  apply(say-underlined, *standard-output*, underline-char, arguments);
end method say-underlined;

define method say-underlined (stream == #f,
			      underline-char :: <character>,
			      #rest arguments)
 => (formatted-string :: <byte-string>,
     underlining :: <byte-string>);
  let formatted-string = apply(say, #f, arguments);
  let underlining = make(<byte-string>,
			 size: formatted-string.size,
			 fill: underline-char);
  values(formatted-string, underlining);
end method say-underlined;

// SAY-NEWLINE is used to display a newline.
//
define method say-newline (stream :: <stream>,
			   #key line-start = "", each-line-start = "",
			   repeat = 1) => ();
  write-line(stream, "");
  for (i from 2 to repeat)
    write-line(stream, each-line-start);
  end for;
  say(stream, line-start);
end method say-newline;

define method say-newline (stream == #t,
			   #key line-start = "", each-line-start = "",
			   repeat = 1) => ();
  say-newline(*standard-output*, line-start: line-start,
	      each-line-start: each-line-start, repeat: repeat);
end method say-newline;

// Does say-newline(#f) make sense?
