module:    Test-Utilities-Implementation
author:    Matthias Hlzl (tc@gauss.muc.de)
copyright: Copyright (C) 1996, 1997, 1998 Matthias Hlzl
           Copyright (C) 1998 Way Forward Technologies
version:   0.04 10 Dec 1998
synopsis:  This Module implements some methods useful for displaying
           and evaluate the results of running tests.

// Copyright.
// =========

// A library for displaying results in a test-suite.
// Copyright (C) 1998 Matthias Hlzl.
// Copyright (C) 1998 Way Forward Technologies.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston,
// MA 02111-1307  USA
//
// If you need to receive this program under another license contact
// the author (tc@gauss.muc.de).

// Bugs.
// ====

// This is a complete rewrite and most of the functionality is either
// not implemented yet or broken.

// The class <TEST-RESULT> holds the result of an individual test.
//
define class <test-result> (<object>)
  slot messages :: <list> = #();
  slot errors-found? :: <boolean> = #f;
end class <test-result>;

define method add-message (msg :: <string>, result :: <test-result>)
 => ();
  result.messages := pair(msg, result.messages);
end method add-message;

// The class <ERROR-RESULT> is the class of errors in the test suite
// that are important and not expected.
//
define class <error-result> (<test-result>)
end class <error-result>;

// The class <EXPECTED-ERROR-RESULT> is the class of important errors
// that are known but not yet fixed.
//
define class <expected-error-result> (<test-result>)
end class <expected-error-result>;

// The class <warning-result> is the class of results that the user
// should be informed about but that are not serious enough to be
// considered errors.
//
define class <warning-result> (<test-result>)
end class <warning-result>;

// The variable *CURRENT-TEST-RESULT* holds the data of the test we
// are currently working on.  The variable *CURRENT-TEST-STATUS* holds
// the status of the current test, i.e., whether it is an error,
// expected-error or warning.
//
define variable *current-test-result* :: <test-result>
  = make(<test-result>);
define variable *current-test-status*
//  :: one-of(#"error", #"expected-error", #"warning", #"normal")
  = #"normal";

// The class <TEST-SECTION> holds the results of one test section.
//
define class <test-section> (<object>)
  slot number-of-errors :: <integer> = 0;
  slot verbosity // :: one-of(#"silent", #"normal", #"verbose", #"gabby")
    = #"normal",
    init-keyword: #"verbosity";
  slot importance :: one-of(#"optional", #"required", #"crucial");
  slot results :: <list> = #();
end class <test-section>;

define method add-result (result :: <test-result>,
			  section :: <test-section>)
 => ();
  section.results := pair(result, section.results);
end method add-result;

define variable *current-test-section* :: <test-section>
  = make(<test-section>);

// The class <TEST-SUITE> holds the result of all tests in the
// test-suite.
//
define class <test-suite> (<object>)
  slot sections :: <list> = #();
  slot verbosity :: one-of(#"silent", #"quiet", #"normal",
			   #"verbose", #"gabby")
    = #"normal",
    init-keyword: #"verbosity";
  slot number-of-errors :: <integer> = 0;
end class <test-suite>;

// PRINT-MESSAGES? returns true for those verbosity-classes for which
// normal messages should be printed.
//
define method print-messages? (verbosity == #"gabby")
  #t;
end method print-messages?;

define method print-messages? (verbosity == #"verbose")
  #t;
end method print-messages?;

define method print-messages? (verbosity == #"normal")
  #t;
end method print-messages?;

define method print-messages? (verbosity == #"quiet")
  #f;
end method print-messages?;

define method print-messages? (verbosity == #"silent")
  #f;
end method print-messages?;

// PRINT-ERROR-MESSAGES returns true for those verbosity-classes for
// which error messages should be printed.  For other verbosity
// classes only the test-suite title and the summary are printed.
//
define method print-error-messages? (verbosity :: <symbol>)
  print-messages?(verbosity);
end method print-error-messages?;

define method print-error-messages? (verbosity == #"quiet")
  #t;
end method print-error-messages?;

// PRINT-WARNING-MESSAGES returns true for those verbosity-classes for
// which warnings are displayed in addition to the normal messages.
//
define method print-warning-messages? (verbosity :: <symbol>)
  print-messages?(verbosity);
end method print-warning-messages?;

define method print-warning-messages? (verbosity == #"normal")
  #f;
end method print-warning-messages?;

// ADD-SECTION adds a <TEST-SECTION> to a <TEST-SUITE>.
//
define method add-section (section :: <test-section>,
			  test-suite :: <test-suite>)
 => ();
  test-suite.sections := pair(section, test-suite.sections);
end method add-section;


// The variable *CURRENT-TEST-SUITE* holds the results of the current
// test-suite.
//
define variable *current-test-suite* :: <test-suite>
  = make(<test-suite>);

// ANNOUNCE-TEST-SUITE created a new global test-suite and prints
// STRING as banner.

define method announce-test-suite(string :: <string>,
				  #key verbosity: v = #"normal")
  *current-test-suite* := make(<test-suite>, verbosity: v);
  say-underlined(#t, '=', string);
end method announce-test-suite;

// ANNOUNCE-SECTION is used to announce a new section of the
// test suite.
//
define method announce-section (string :: <string>,
				#key verbosity: v = #f)
  *current-test-section*
    := make(<test-section>,
	    verbosity: v | *current-test-suite*.verbosity);
  add-section(*current-test-section*, *current-test-suite*);
  if (print-error-messages?(*current-test-section*.verbosity))
    say-underlined(#t, '-', string);
  end if;
end method announce-section;

// ANNOUNCE is used to annouce that a new test starts.
//
define method announce (string :: <string>, #rest arguments)
  *current-test-result* := make(<test-result>);
  *current-test-status* := #"normal";
  add-result(*current-test-result*, *current-test-section*);
  if (print-messages?(*current-test-section*.verbosity))
    apply(say-dots, #t, string, arguments);
  end if;
end method announce;

// COUNT-ERROR adds one to all error counters.
//
define function count-error () => ();
  *current-test-section*.number-of-errors
    := *current-test-section*.number-of-errors + 1;
  *current-test-suite*.number-of-errors
    := *current-test-suite*.number-of-errors + 1;
end function count-error;

// REPORT-ERROR is the standard way to report errors.
//
define method report-error (#rest arguments) => ();
  *current-test-status* := #"error";
  *current-test-result*.errors-found? := #t;
  count-error();
  unless (arguments.empty?)
    let message-string = apply(say, #f, arguments);
    add-message(message-string, *current-test-result*);
    if (print-error-messages?(*current-test-section*.verbosity))
      if (print-messages?(*current-test-section*.verbosity))
	say-newline(#t, line-start: "  ");
      end if;
      say-dots(#t, message-string);
    end if;
  end unless;
end method report-error;

// REPORT-EXPECTED-ERROR is used to report errors that are known but
// not fixed..
//
define method report-expected-error (#rest arguments) => ();
  if (*current-test-status* = #"normal"
	| *current-test-status* = #"warning")
    *current-test-status* := #"expected-error";
  end if;
  unless (arguments.empty?)
    let message-string = apply(say, #f, arguments);
    add-message(message-string, *current-test-result*);
    if (print-messages?(*current-test-section*.verbosity))
      say-newline(#t, line-start: "  ");
      say(#t, "Known Error: ");
      say-dots(#t, message-string);
    end if;
  end unless;
end method report-expected-error;

// REPORT-WARNING is used to report unexpected situations that are not
// errors.
//
define method report-warning (#rest arguments) => ();
  if (*current-test-status* = #"normal")
    *current-test-status* := #"warning";
  end if;
  unless (arguments.empty?)
    let message-string = apply(say, #f, arguments);
    add-message(message-string, *current-test-result*);
    if (print-warning-messages?(*current-test-section*.verbosity))
      say-newline(#t, line-start: "  ");
      say(#t, "Warning: ");
      say-dots(#t, message-string);
    end if;
  end unless;
end method report-warning;

// SAY-DONE is called after each test and produces the correct
// end-of-test message.
//
define method say-done()
  if (*current-test-result*.errors-found? = #f)
    if (print-messages?(*current-test-section*.verbosity))
    say-line(#t, "done.");
    end;
  else
    if (print-error-messages?(*current-test-section*.verbosity))
      say-line(#t, "ERROR!");
    end if;
  end if;
end method say-done;

// SAY-INTERMEDIATE-RESULTS is used to report the number of
// errors after a part of the test suite.
//
define method say-intermediate-results()
  if (print-messages?(*current-test-section*.verbosity))
    if (*current-test-section*.number-of-errors = 0)
      say-line(#t, "All tests in this section completed without errors.");
    elseif (*current-test-section*.number-of-errors = 1)
      say-line(#t, "There was one error in this section!");
    else
      say-line(#t, "There were %D errors in this section!",
	  *current-test-section*.number-of-errors);
    end if;
    say-newline(#t);
  elseif (print-error-messages?(*current-test-section*.verbosity))
    say-newline(#t);
  end if;
end method say-intermediate-results;

// SAY-RESULTS is used to report a summary of all tests.
//
define method say-results()
//  if (print-messages?(*current-test-section*.verbosity))
//    say-newline(#t);
//  end if;
  
  if (*current-test-suite*.number-of-errors = 0)
    say-line(#t, "All tests completed without errors.");
  elseif (*current-test-suite*.number-of-errors = 1)
    say-line(#t, "There was one error in the test-suite!");
  else
    say(#t, "There were %D errors in the test-suite!",
	*current-test-suite*.number-of-errors);
  end if;
  say-newline(#t, repeat: 2);
end method say-results;
